from PyQt5 import uic
from PyQt5.QtWidgets import QLabel, QLineEdit, QWidget

from nanoqnt.view import VIEW_FOLDER


class ChannelSelector(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(str(VIEW_FOLDER / 'GUI' / 'channel_selector.ui'), self)
        self.update_channels(0)
        self.channel_box.currentIndexChanged.connect(self.update_channels)
        self.channels = []
        self.accept_button.clicked.connect(self.close)

    def update_channels(self, number):
        self.channels = []
        layout = self.channels_widget.layout()
        for i in range(layout.rowCount()):
            layout.removeRow(0)

        for i in range(number+1):
            self.channels.append(QLineEdit(f"Name {i}"))
            layout.addRow(QLabel(f"Channel {i}"), self.channels[-1])



from functools import wraps

from .exceptions import NoData


def data_not_none(func):
    @wraps
    def func(*args, **kwargs):
        if getattr(args[0], 'data', None) is None:
            raise NoData('Data not yet defined')
        return func(args, kwargs)
    return func


from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
import pyqtgraph as pg
from nanoqnt.view import VIEW_FOLDER


class ParticleWidget(QWidget):
    """Wrapper class to display image data with few signals to update the frame. It is meant to handle also
    multi-color data"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        uic.loadUi(str(VIEW_FOLDER / 'GUI' / 'particles_widget.ui'), self)
        self.image = pg.ImageView()
        self.image.setPredefinedGradient('thermal')

        self.circles = []

        plot_layout = self.video_widget.layout()
        plot_layout.addWidget(self.image)

    def autoRange(self):
        """ Wrapper to the default PyQtGraph auto range to see if it integrates better with the overall UI"""
        # self.image.view.setXRange(0, 2000)
        self.image.view.setRange(xRange=(0, 2000), yRange=(0,2000))

    def update_image(self, image, circles=None, auto_range=False, auto_levels=False):
        for r in self.circles:
            self.image.removeItem(r)
            self.circles = []

        self.image.setImage(image, autoRange=auto_range, autoLevels=auto_levels)

        if circles is not None:

            self.circles = circles

            for p in self.circles:
                try:  # Very dangerous, but this is a feature fixed by PyQtGraph only on version > 0.13.2
                    p.removeHandle(0)
                except:
                    pass
                self.image.addItem(p)
            self.found_particles_line.setText(str(len(circles)))

        if auto_range:
            self.autoRange()
        if auto_levels:
            self.image.autoLevels()

    def removeItem(self, r):
        self.image.removeItem(r)

    def setImage(self, *args, **kwarks):
        self.image.setImage(*args, **kwarks)

    def addItem(self, p):
        self.image.addItem(p)

    def autoLevels(self):
        self.image.autoLevels()


import sys

from PyQt5.QtWidgets import QApplication

from nanoqnt.model.stage_control import StageControl
from nanoqnt.view.stage_control import StageControlWindow


def start_stage_control():
    model = StageControl()
    app = QApplication([])
    window = StageControlWindow(model)
    window.show()
    sys.exit(app.exec())


if __name__ == '__main__':
    start_stage_control()


import os
from multiprocessing import freeze_support

import sys

from PyQt5 import QtGui
from PyQt5.QtWidgets import QApplication

from nanoqnt.model.analyze_nanoqnt import AnalyzeNanoQNT
from nanoqnt.view.main_window import NanoQNTMainWindow

try:
    from ctypes import windll  # Only exists on Windows.
    myappid = 'Dispertech.NanoQNT.Analysis.0.2.0'
    windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
except ImportError:
    pass

def start_nanoqnt():
    basedir = os.path.dirname(__file__)
    model = AnalyzeNanoQNT()
    app = QApplication([])
    app.setWindowIcon(QtGui.QIcon(os.path.join(basedir, 'dispertech-logo.ico')))
    main_window = NanoQNTMainWindow(model)
    main_window.show()
    app.exec()
    sys.exit()


if __name__ == '__main__':
    freeze_support()
    start_nanoqnt()



import time
from pathlib import Path

import matplotlib.pyplot as plt
from PyQt5 import uic
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QFileDialog, QMainWindow, QMessageBox
import pyqtgraph as pg

from nanoqnt.view import VIEW_FOLDER
from nanoqnt.view.channel_selector import ChannelSelector
from nanoqnt.view.particle_widget import ParticleWidget

home_path = Path.home()


class NanoQNTMainWindow(QMainWindow):
    def __init__(self, model=None):
        """
        :param AnalyzeNanoQNT model: Model used to analyze the data
        """
        super().__init__(parent=None)
        uic.loadUi(str(VIEW_FOLDER / 'GUI' / 'data_exploration.ui'), self)
        self.setWindowTitle('NanoQNT Analysis')

        self.analyze_model = model

        self.action_open.triggered.connect(self.open)
        self.action_mass_histogram.triggered.connect(self.histogram_mass)
        self.action_data_frame.triggered.connect(self.data_over_frame)
        self.action_data_plots.triggered.connect(self.total_histograms)
        self.action_export_all_data.triggered.connect(self.export_all_data)
        self.action_save_data.triggered.connect(self.export_data)
        self.action_calculate_bkg.triggered.connect(self.calculate_background)
        self.action_show_bkg.triggered.connect(self.show_background)

        self.frame_slider.valueChanged.connect(self.update_slider)
        self.step_size_line.editingFinished.connect(self.calculate_concentration)
        self.find_particles_check.toggled.connect(self.update_images)
        self.calculate_button.clicked.connect(self.calculate_all)
        self.action_about.triggered.connect(self.show_about)

        self.circles = []
        self.curr_index = 0

        self.channel_widget = ChannelSelector()
        self.channel_widget.accept_button.clicked.connect(self.channels_updated)

        self.nanoqnt_images = []

    def open(self):
        last_dir = self.analyze_model.metadata.get('last_dir', home_path)
        file = QFileDialog.getOpenFileName(self, 'Open Data', str(last_dir), filter='*.tif')[0]
        if file != '':
            file = Path(file)
        else:
            return

        self.channel_widget.show()

        self.analyze_model.open(str(file))
        self.setWindowTitle(f'NanoQNT Analysis: {file.name}')

    def update_slider(self, index):
        self.curr_index = index - 1
        self.frame_num.setText(str(index))
        self.update_images()

    def update_images(self, auto_range=False, auto_levels=False):
        if self.analyze_model.data is None:
            return


        for i, channel_image in enumerate(self.nanoqnt_images):
            index = self.curr_index*len(self.nanoqnt_images) + i
            circles = None
            if self.find_particles_check.isChecked():
                diameter = int(channel_image.size_line.text())
                minmass = int(channel_image.min_mass_line.text())
                method = self.engine_box.currentText()

                self.analyze_model.find_particles(index, diameter, minmass, method=method)
                particles_df = self.analyze_model.particle_df[index]
                circles = [pg.CircleROI([p[1] - 12, p[0] - 12], [25, 25],
                              pen=pg.mkPen('r', width=2),
                              movable=False,
                              resizable=False,
                              rotatable=False,
                              removable=False,
                              )
                    for p in zip(particles_df['x'], particles_df['y'])]
            channel_image.update_image(self.analyze_model.data[index], circles, auto_range=auto_range,
                                       auto_levels=auto_levels)

    def histogram_mass(self):
        index = self.curr_index
        diameter = int(self.size_line.text())
        minmass = int(self.min_mass_line.text())

        self.analyze_model.find_particles(index, diameter, minmass)

        fig, ax = plt.subplots(1)
        ax.hist(self.analyze_model.particle_df[index]['mass'], bins=50)
        fig.suptitle(f'Mass Histogram Frame Number: {self.curr_index}')
        fig.show()

    def calculate_all(self):
        QApplication.setOverrideCursor(Qt.BusyCursor)
        diameter = [int(channel_image.size_line.text()) for channel_image in self.nanoqnt_images]
        minmass = [int(channel_image.min_mass_line.text()) for channel_image in self.nanoqnt_images]
        engine = self.engine_box.currentText()
        frames_no = [int(self.start_frame_line.text())-1, int(self.end_frame_line.text())]
        self.analyze_model.find_all_particles(frames_no, diameter, minmass, engine)
        search_radius = int(self.search_radius_line.text())
        min_frames = int(self.min_frames_line.text())
        self.analyze_model.link_particles(search_radius, memory=0, min_frames=min_frames)
        self.analyze_model.calculate_intensities()
        self.analyze_model.multi_color_link()
        self.calculate_concentration()
        QApplication.setOverrideCursor(Qt.ArrowCursor)

    def data_over_frame(self):
        t1 = self.analyze_model.linked_particles

        fig, ax = plt.subplots(3, sharex=True)
        ax[0].plot(t1['ecc'].groupby('frame').mean())
        ax[0].set_ylabel('Eccentricity')
        ax[1].plot(t1['size'].groupby('frame').mean())
        ax[1].set_ylabel('Size')
        ax[2].plot(t1['mass'].groupby('frame').mean())
        ax[2].set_ylabel('Mass')
        ax[2].set_xlabel('Frame Num')
        fig.show()

    def total_histograms(self):
        if self.analyze_model.linked_particles is None:
            print('No particles to plot!')
            msg = QMessageBox.warning(self, "Run Analysis First", "Before being able to show the histogram of "
                                                                      "all "
                                                                      "the particles, you must run the analysis using "
                                                                      "the calculate all button.",
                                      QMessageBox.Ok)
            return

        for channel in range(self.analyze_model.num_channels):
            t1 = self.analyze_model.summary_data[channel]
            fig1, ax1 = plt.subplots(1)
            t1[f'i_{channel}'].hist(bins=100, ax=ax1)
            ax1.set_xlabel('Total Intensity (counts)')
            ax1.set_title(f'Channel: {channel}')
            fig1.show()

            for channel_2 in range(channel+1, self.analyze_model.num_channels):
                fig, ax = plt.subplots(1) #, figsize=(10, 8))
                ax.plot(self.analyze_model.summary_data[channel][f'i_{channel}'],
                        self.analyze_model.summary_data[channel][f'i_{channel_2}'],
                        'o',
                        alpha=0.1)
                ax.set_xlabel(self.channel_names[channel], fontsize=14)
                ax.set_ylabel(self.channel_names[channel_2], fontsize=14)
                plt.gca().tick_params(labelsize=12)
                # ax.set_yscale('log')
                # ax.set_xscale('log')
                fig.show()

    def calculate_concentration(self):
        step_size = float(self.step_size_line.text())
        self.analyze_model.calculate_concentration(step_size)
        for i in range(self.analyze_model.num_channels):
            self.nanoqnt_images[i].concentration_line.setText(f"{self.analyze_model.concentration[i]:.2E}")
            self.nanoqnt_images[i].total_particles_line.setText(f"{self.analyze_model.total_num_particles[i]}")

    def export_all_data(self):
        path = self.analyze_model.filename.parent
        filename = QFileDialog.getSaveFileName(self, 'Save to file', str(path), '*.csv')[0]

        if filename is not None:
            self.analyze_model.save_all_data(filename)

    def export_data(self):
        path = self.analyze_model.filename.parent
        filename = QFileDialog.getSaveFileName(self, 'Save to file', str(path), '*.csv')[0]

        if filename is not None:
            self.analyze_model.save_data(filename)

    def channels_updated(self):
        self.analyze_model.num_channels = self.channel_widget.channel_box.currentIndex()+1
        self.channel_names = [line.text() for line in self.channel_widget.channels]


        self.nanoqnt_images = [ParticleWidget() for _ in range(self.analyze_model.num_channels)]

        plot_layout = self.video_widget.layout()
        for i in reversed(range(plot_layout.count())):
            plot_layout.itemAt(i).widget().setParent(None)

        if self.analyze_model.num_channels == 1:
            plot_layout.addWidget(self.nanoqnt_images[0], 0, 0)
        elif self.analyze_model.num_channels == 2:
            plot_layout.addWidget(self.nanoqnt_images[0], 0, 0)
            plot_layout.addWidget(self.nanoqnt_images[1], 0, 1)
        elif self.analyze_model.num_channels == 3:
            plot_layout.addWidget(self.nanoqnt_images[0], 0, 0)
            plot_layout.addWidget(self.nanoqnt_images[1], 0, 1)
            plot_layout.addWidget(self.nanoqnt_images[2], 1, 0)
        elif self.analyze_model.num_channels == 4:
            plot_layout.addWidget(self.nanoqnt_images[0], 0, 0)
            plot_layout.addWidget(self.nanoqnt_images[1], 0, 1)
            plot_layout.addWidget(self.nanoqnt_images[2], 1, 0)
            plot_layout.addWidget(self.nanoqnt_images[3], 1, 1)

            # print(plot_layout.itemAt(i).widget().image.view.state)

        if self.analyze_model.num_channels > 1:
            for i, plot in enumerate(self.nanoqnt_images[:-1]):
                view = plot.image.view
                view.linkView(self.nanoqnt_images[i+1].image.view.XAxis, self.nanoqnt_images[i+1].image.view)
                view.linkView(self.nanoqnt_images[i+1].image.view.YAxis, self.nanoqnt_images[i+1].image.view)

        self.frame_slider.setRange(1, int(self.analyze_model.num_frames/self.analyze_model.num_channels))
        self.frame_slider.setEnabled(True)
        self.start_frame_line.setText("1")
        self.end_frame_line.setText(str(int(self.analyze_model.num_frames/self.analyze_model.num_channels)))
        self.update_slider(1)

        self.update_images(auto_range=True, auto_levels=True)
        for i in reversed(range(plot_layout.count())):
            plot_layout.itemAt(i).widget().autoRange()
            plot_layout.itemAt(i).widget().min_mass_line.editingFinished.connect(self.update_images)
            plot_layout.itemAt(i).widget().size_line.editingFinished.connect(self.update_images)

    def calculate_background(self):
        print('Calculating background')
        QApplication.setOverrideCursor(Qt.BusyCursor)
        frames_no = [int(self.start_frame_line.text())-1, int(self.end_frame_line.text())]
        self.analyze_model.calculate_background(frames_no=frames_no)
        QApplication.setOverrideCursor(Qt.ArrowCursor)
        print('Done calculating background')

    def show_background(self):
        fig, ax = plt.subplots(1)
        ax.imshow(self.analyze_model.bkg)
        fig.suptitle(f'Mass Histogram Frame Number: {self.curr_index}')
        fig.show()

    def show_about(self):
        QMessageBox.about(self, 'About NanoQNTPy', 'Beta version 0.1')


from time import sleep

import pyvisa

rm = pyvisa.ResourceManager('@py')


class StageControl:
    def __init__(self):
        self.scan_running = False
        self.stage = rm.open_resource('ASRL8::INSTR')
        self.arduino = rm.open_resource('ASRL7::INSTR')
        self.stage.baud_rate = 115200
        self.arduino.baud_rate = 115200
        self.arduino.write_termination = '\r'
        self.stage.read_termination = '\n'
        self.stage.write_termination = '\r'
        sleep(2)
        self.arduino.timeout = 5000
        self.stage.timeout = 5000

    def live(self):
        print(self.arduino.query('live').strip())

    def stop_live(self):
        print(self.arduino.query('stop').strip())

    def snap(self):
        print(self.arduino.query('snap').strip())

    def move(self, position):
        print(self.stage.query(f'GT z{position}').strip())

    def prepare_scan(self, start):
        self.stop_live()
        print('Move to start')
        self.stage.query(f'GT z{start}')
        while self.stage.query('RZ z') == '$RS z1':
            sleep(0.01)

    def scan(self, start, end, step_size):
        self.scan_running = True

        if (end-start)/step_size < 0:
            step_size = -step_size

        for p in range(start, end, step_size):
            self.stage.query(f'GT z{p}')
            while self.stage.query('RS z') == '$RS z1':
                sleep(0.001)
            sleep(0.1)
            self.arduino.query('snap')

        print('Done')


import numpy as np
import skimage
import yaml
from pathlib import Path

import pandas as pd
import pims as pims
import trackpy as tp
from scipy.spatial import KDTree
from skimage import morphology
from skimage.measure import label, regionprops, regionprops_table
from trackpy.find import where_close
from trackpy.utils import pandas_concat
import scipy.optimize as opt

from nanoqnt.util import twoD_Gaussian


class AnalyzeNanoQNT:
    def __init__(self):
        self.filename = None
        self.data = None
        self.metadata = {}
        self.particle_df = []
        self.particles = pd.DataFrame()
        self.num_frames = 0
        self.find_settings = {}
        self.filename = Path.home()
        self.concentration = 0  # pcles/ml
        self.linked_particles = None
        self.total_num_particles = 0
        self.summarized_particles = None

        self.ignore_edge_pixels = 8

        self.num_channels = 1
        self.channels_data = {}
        self.bkg = None

    def open(self, filename):
        """ Opens the file with the data. Normally a multi-tiff file, but it can be adapted to other formats.

        :param filename:
        """
        self.filename = Path(filename)
        self.metadata['filename'] = filename
        self.metadata['last_dir'] = self.filename.parent
        self.data = pims.open(filename)
        self.num_frames = len(self.data)  # This is really the number of frames in the data, not the number of
        # multi-channel frames

        self.particle_df = [pd.DataFrame() for _ in range(len(self.data))]
        self.particles = pd.DataFrame()
        self.find_settings = {}
        self.linked_particles = None
        self.summarized_particles = None
        self.total_num_particles = 0
        self.channels_data = {}

    def find_particles(self, frame_num, diameter, minmass, method='mask'):
        """ Finds particles in a given frame and stores the data frame in a list indexed by the frame number.

        :param int frame_num: The number of the frame (0-indexed) to analyze
        :param int diameter: The expected diameter of the particles. Must be an odd number and is passed directly to
        Trackpy
        :param int minmass: Filter to select bright enough spots. The parameter is passed directly to Trackpy.
        :param str method: Either select trackpy or mask. Default is mask.
        """
        if method == 'python' or method == 'numba':
            if frame_num in self.find_settings:
                settings = self.find_settings[frame_num]
                if diameter == settings['diameter'] and minmass == settings['minmass']:
                    return self.particle_df[frame_num]
            self.find_settings[frame_num] = {
                'diameter': diameter,
                'minmass': minmass,
                }
            df = tp.locate(self.data[frame_num], diameter, minmass=minmass, characterize=False)
            df = df.loc[df['y'] >= self.ignore_edge_pixels]
            self.particle_df[frame_num] = df
            return df

        elif method == 'mask':
            mask = self.data[frame_num] > minmass
            mask = morphology.remove_small_objects(mask, diameter)
            label_img = label(mask)
            props = regionprops_table(label_img, intensity_image=np.array(self.data[frame_num]), properties=('centroid',
                                                                                                             'intensity_mean',))
            df = pd.DataFrame(props).rename(columns={'centroid-0': 'y', 'centroid-1': 'x', 'intensity_mean': 'mass'})
            to_drop = where_close(df[['x', 'y']], separation=diameter * 2, intensity=df['mass'])
            df = df.drop(to_drop)
            df = df.loc[df['y'] >= self.ignore_edge_pixels]
            self.particle_df[frame_num] = df
            return df

    def find_all_particles(self, frames_no, diameter, minmass, method='python'):
        """ Finds all the particles in a given range of frames. If the data has multiple channels, then it can handle
        different settings for each one. In that case, diameter and minmass must be iterables and in the appropriate
        order, i.e., the first setting will be for the first channel, etc.

        The data is structured as a dictionary, where each channel is a different key, and then the localizations are
        stored as a pandas dataframe.

        :param list frames_no:
        :param tuple diameter:
        :param tuple minmass:
        :param str method:
        :return dict: dictionary of particles found in each channel
        """
        try:
            if len(diameter) != self.num_channels:
                diameter = self.num_channels * [diameter]
        except TypeError:
            diameter = self.num_channels * [diameter]
        try:
            if len(minmass) != self.num_channels:
                minmass = self.num_channels * [minmass]
        except TypeError:
            minmass = self.num_channels * [minmass]

        if len(frames_no) == 1:
            frames_no = (0, self.num_frames)

        self.particles = {}

        self.metadata.update({
            'frames': frames_no,
            'diameter': diameter,
            'minmass': minmass,
            'engine': method,
            })

        if method == 'mask':
            for channel in range(self.num_channels):
                all_features = []
                for frame in range(frames_no[0], frames_no[1]):
                    real_frame = frame*self.num_channels + channel
                    df = self.find_particles(real_frame, diameter[channel], minmass[channel], method='mask')
                    df['frame'] = frame
                    all_features.append(df)
                self.particles[channel] = pandas_concat(all_features).reset_index(drop=True)
            return self.particles

        ## For the python or numba methods, only 1-channel data is available
        if self.num_channels > 1:
            raise Exception("Trackpy-based methods only work with 1-channel data for the time being")

        self.particles[0] = tp.batch(self.data[frames_no[0]:frames_no[1]],
                                               diameter[0],
                                               minmass=minmass[0],
                                               characterize=False,
                                               preprocess=False,
                                               processes='auto',
                                               engine=method)
        self.particles[0]['intensity'] = np.nan

        return self.particles

    def link_particles(self, search_radius, memory=0, min_frames=0):
        self.linked_particles = {}
        for channel in range(self.num_channels):
            self.linked_particles[channel] = tp.link(self.particles[channel], search_radius, memory=memory)
            self.linked_particles[channel] = tp.filter_stubs(self.linked_particles[channel], min_frames)

        self.metadata.update({
                                 'search_radius': search_radius,
                                 'min_frames': min_frames,
                                 'memory': memory
                                 })

    def calculate_intensities(self):
        """Calculates the intensities of the linked particles assuming they were the product of the mask method.
        The methods that rely on the trackpy algorithm already include the intensity.
        """
        fit_size = 10
        x_fit = np.linspace(0, 2 * fit_size, 2 * fit_size + 1)
        y_fit = np.linspace(0, 2 * fit_size, 2 * fit_size + 1)
        x_fit, y_fit = np.meshgrid(x_fit, y_fit)
        initial_guess = (200, fit_size, fit_size, 2, 2, 200)

        self.summary_data = {}
        intensity_columns = [f'i_{channel}' for channel in range(self.num_channels)]
        for channel in range(self.num_channels):
            self.summary_data[channel] = pd.DataFrame(columns=['particle', 'frame', 'x', 'y'] + intensity_columns)

            for p in self.linked_particles[channel]['particle'].unique():
                pcle = self.linked_particles[channel].loc[self.linked_particles[channel]['particle'] == p]
                ix = pcle['mass'].idxmax()
                x = int(pcle.loc[ix]['y'])
                y = int(pcle.loc[ix]['x'])
                frame = int(pcle.loc[ix]['frame'])
                # TODO: This is hardcoded and dangerous. The margin used to drop particles must be properly considered
                if x < 20 or x > 2048 - 20 or y < 20 or y > 2048 - 20:
                    continue

                # try:
                #     f = frame*self.num_channels + channel
                #     popt, pcov = opt.curve_fit(twoD_Gaussian, (x_fit, y_fit),
                #                                np.array(self.data[f][x - fit_size:x + fit_size + 1,
                #                                         y - fit_size:y + fit_size + 1]).reshape(-1),
                #                                p0=initial_guess)
                # except Exception as e:
                #     print(e)
                #     continue

                df = pd.DataFrame.from_dict({
                    'particle': [p],
                    'frame': [pcle.loc[ix]['frame']],
                    'x': [pcle.loc[ix]['x']],
                    'y': [pcle.loc[ix]['y']],
                    f'i_{channel}': [pcle['mass'].max()]})
                self.summary_data[channel] = pd.concat((self.summary_data[channel], df), ignore_index=True)
            self.summary_data[channel] = self.summary_data[channel].reset_index(drop=True)

    def multi_color_link(self):
        """ Find the same particles in every channel and add the intensity information on each """
        kd_trees = {}
        for channel in range(self.num_channels):
            kd_trees[channel] = KDTree(np.array(self.summary_data[channel][['x', 'y']]))

        for channel in range(self.num_channels):
            for channel_2 in range(channel+1, self.num_channels):
                for ix, p in self.summary_data[channel].iterrows():
                    dist, ind = kd_trees[channel_2].query((p['x'], p['y']), p=2, distance_upper_bound=15)
                    if ind == kd_trees[channel_2].n:
                        continue
                    if abs(self.summary_data[channel_2].loc[ind]['frame'] - p['frame']) > 5:
                        print('Frames too far apart with ', ix, p['x'], p['y'], p['frame'])
                        continue
                    self.summary_data[channel].loc[ix, f'i_{channel_2}'] = \
                        self.summary_data[channel_2].loc[ind][f'i_{channel_2}']
                    self.summary_data[channel_2].loc[ind, f'i_{channel}'] = p[f'i_{channel}']
    def save_all_data(self, filename):
        self.linked_particles.to_csv(str(filename))

    def save_data(self, filename):
        """ Saves only the most important information to the file: particle and intensity information.
        """
        if not type(filename) is Path:
            filename = Path(filename)

        for i in range(self.num_channels):
            stem = filename.stem
            file_name = filename.with_stem(f'{stem}_{i}')
            self.summary_data[i].to_csv(str(file_name))

        with open(filename.with_suffix('.yml'), 'w') as f:
            yaml.dump(self.metadata, f)

    def sub_pix_bias(self, index):
        f = self.particle_df[index]
        return f[['x', 'y']].applymap(lambda x: x % 1)

    def calculate_concentration(self, step_size=5):
        self.total_num_particles = {}
        self.concentration = {}
        if self.linked_particles is None:
            return
        fraction_sensor_used = (2048 - self.ignore_edge_pixels) / 2048
        field_of_view = fraction_sensor_used * (13.3 / 20) ** 2  # (Sensor size/magnification) # mm2
        for channel in range(self.num_channels):
            frames = self.linked_particles[channel]['frame'].max() - self.linked_particles[channel]['frame'].min() + 1
            volume = field_of_view * frames * step_size / 1000  # In mm3 (step_size in microns)
            volume = volume / 1000  # in ml
            num_particles = len(self.linked_particles[channel].groupby('particle').sum())

            self.total_num_particles[channel] = num_particles
            self.concentration[channel] = num_particles / volume

        self.metadata.update({'concentration': self.concentration})

    def calculate_background(self, frames_no=None, num_frames=10, sigma=200):
        if frames_no is None:
            average = np.mean(self.data[:num_frames], 0)
        else:
            average = np.mean(self.data[frames_no[0]:frames_no[1]], 0)
        self.bkg = skimage.filters.gaussian(average, sigma=sigma)



import numpy as np


def twoD_Gaussian(xy, amplitude, xo, yo, sigma_x, sigma_y, offset):
    """ Used for fitting the PSF in 2-D. """
    x, y = xy
    xo = float(xo)
    yo = float(yo)
    a = 1/(2*sigma_x**2)
    c = 1/(2*sigma_y**2)
    g = offset + amplitude*np.exp(-(a*((x-xo)**2) + c*((y-yo)**2)))
    return g.ravel()


from threading import Thread

from PyQt5 import uic
from PyQt5.QtWidgets import QMainWindow, QMessageBox

from nanoqnt.view import VIEW_FOLDER


class StageControlWindow(QMainWindow):
    def __init__(self, model=None):
        """
        :param StageControl model: Model used to control the stage
        """
        super().__init__(parent=None)
        uic.loadUi(str(VIEW_FOLDER/'GUI'/'scan_control.ui'), self)
        self.model = model

        self.start_live_button.clicked.connect(self.model.live)
        self.stop_live_button.clicked.connect(self.model.stop_live)
        self.snap_button.clicked.connect(self.model.snap)
        self.scan_button.clicked.connect(self.scan)
        self.goto_button.clicked.connect(self.move)

    def scan(self):
        start = int(self.start_line.text())
        end = int(self.stop_line.text())
        step_size = int(self.step_line.text())
        self.model.prepare_scan(start)

        self.msgbox = QMessageBox()
        self.msgbox.setText(f'Open Multi-D Acq. in Micromanager and set the count to {int((end-start)/step_size)}')
        self.msgbox.exec()

        self.scan_thread = Thread(
            target=self.model.scan, args=(start, end, step_size)
            )
        self.scan_thread.start()

    def move(self):
        goto = int(self.goto_line.text())
        self.model.move(goto)